# Mercurial projects in Heptapod

The primary objective of Heptapod is to add Mercurial support to GitLab.

## Version Control System (VCS) types

Each project has a `vcs_type` field to describe the VCS in use. Administrators
can select which of these are allowed in the application settings
(Admin Area > Settings > General > Visibility and access controls).


We are currently using two distinct VCS types for Mercurial projects, to support our
effort towards "native" handling of Mercurial.

### "Mercurial" VCS type

Value of the `vcs_type` field: `hg`.

This has been the default VCS type for new Projects since Heptapod 0.25 and was displayed
as "Mercurial (native)" before Heptapod 0.28.

Projects created right now under this mode should not need heavy and risky migrations
in the future. They are often referred to as "native Mercurial projects" in text resources,
such as documentation and issues.

The native Mercurial projects are exposed to the various GitLab components via the HGitaly
service. As of Heptapod 0.28, content is still converted to Git by default, but this can be
disabled by feature flags.

There are four stages of development:

+ HGitaly1: references and commits are served via HGitaly, but files and directories are still read
  from the Git repository. This was the norm before Heptapod 0.28.
+ HGitaly2 (the default since Heptapod 0.28, still undergoing evaluation via feature flags):
  all Mercurial content is served via HGitaly. Further, the conversion to auxiliary Git
  repository can also be disabled by feature flags and is mostly useful only to be able to get
  back to the HGitaly1 stage.
+ HGitaly3 (development not started): all read and write operations, including creation and
  maintenance tasks happen through HGitaly. The web application does not need to access
  the repositories directly, and the HGitaly servers can be remote.

More insight on these different ways to handle Mercurial is available
[on the Heptapod web site](https://heptapod.net/the-road-to-fully-native-mercurial-in-heptapod.md#the-road-to-fully-native-mercurial-in-heptapod)

### "Mercurial (legacy with Git conversion)" VCS type

Value of the `vcs_type` field: `hg_git`.

This VCS type was disabled by default for Project creation in Heptapod 0.25, but can
be enabled again in the application setttings.

This is the historical way to handle Mercurial repositories, exposing to the rest of
GitLab by maintaining a conversion to Git. Even though the server side Mercurial
repository is the single source of truth, this has many drawbacks.

The Rails (web) application accesses content for display through Gitaly, but also needs
to perform operations (such as merges) on the local file system. Hence the Gitaly server
*must* be local.

This does not prevent using also remote Gitaly servers, as long as they serve Git
projects only.

## Managing the fully native modes

Warning: as of Heptapod 0.28, this is experimental, but you can get back to
the HGitaly1 stage easily.

In all cases, only native Mercurial projects are affected by this.

#### The feature flags

- `hg_fully_native`: switches on the fully native mode. Defaults to `true`
  since Heptapod 0.28.
- `hgitaly1_fallback`: make a given project still use the HGitaly1 mode, even
  if the whole instance is using the fully native mode
- `hg_without_git` (for fully native projects only): prevents conversion to Git
  unless the Project explicitely requires it, e.g, to mirror-push to a Git
  server
- `hg_with_git_fallback`: make a given "fully native" project still convert
  content to Git, even if `hg_without_git` is enabled for the whole instance.

#### For the whole instance

To have all native Mercurial projects get back to the HGitaly1 mode of operation,
do this under the [Rails console]:

```ruby
Feature.disable(:hg_fully_native)
```

To enable it again:

```ruby
Feature.enable(:hg_fully_native)
```

It is not possible to use `disable()` for a single project. To exclude a given project
from the fully native mode, do:

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.enable(:hgitaly1_fallback, project)
```

Of course, you are free to use other means to retrieve a `Project` instance, using
the `ActiveRecord` API (the Object Relational Mapping in use).

To stop conversion to Git for "fully native" Mercurial projects, do:

```ruby
Feature.enable?(:hg_without_git)
```

To keep on converting for a given Project even if `hg_without_git` is
globally enabled (assuming no changeset has been pushed in the mode
without Git yet, otherwise see the dedicated section below):

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.enable?(:hg_with_git_fallback, project)
```

#### For a given project

##### Introspection

To know whether a given project is running under the fully native mode, retrieve its
details using the [Projects API](../api/projects.md).

- if `vcs_type` is `hg` and `hgitaly1` is `false`, the project is running under the fully
native mode.
- if `hgitaly1` is `true`, the project is running the HGitaly1 mode of operation
- if `vcs_type` is not `hg`, this is not a native Mercurial repository, the value of
  `hgitaly1` is meaningless.

##### Using the Rest API

This is perhaps the simplest. Use the [Features flags API](../api/features.md#set-or-create-a-feature)
with this data:

- `name`: `hg_fully_native`
- `value`: `true` or `false` whether you want to enable or disable
- `project`: full web path to the project

Note: disabling the fully native mode this way does not work if it is enabled for the
whole instance. Enable the `hgitaly1_fallback` feature flag instead.

Similarly, to try without the conversion to Git (for fully native projects),
use the API to set `hg_without_git` to `true` on the project.


##### Under the [Rails console]:

To get back to HGitaly1 mode of operation:

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.disable(:hg_fully_native, project)
```

To activate the mode without Git conversion on a single project:

```ruby
project = Project.find(project_id)  # project_id is the numeric ID,
Feature.enable(:hg_without_git, project)
```

#### How to resume Git conversion if it has been stopped by feature flags

If you want to *resume* conversion to Git and it has been
stopped for some time, you need to

- (recommended) stop accepting changes by archiving the relevant
  Projects
- run `hg hpd-git-resync` for the given projects, following
  [how to run a hg command on the server](https://heptapod.net/pages/faq.html#hg-command)
- flip the feature flags (either the global one or the per-project) as
  explained above
- unarchive the project if you did archive it.


[Rails console]: operations/rails_console.md




